from upyog.ml.imports import *
from upyog.os import *


__all__ = ["to_np", "to_torch", "save_tensor_to_json"]


def get_num_gpus(gpus: Union[List[int], int]) -> Union[None, int]:
    if is_platform_macos():
        return None
    if gpus == -1:
        return torch.cuda.device_count()
    else:
        assert isinstance(gpus, list)
        return len(gpus)


def to_np(tensor: Tensor, dtype=np.float32):
    if isinstance(tensor, np.ndarray): return tensor
    return tensor.detach().cpu().numpy().astype(dtype)


def to_torch(x) -> Tensor:
    assert isinstance(x, (np.ndarray, torch.Tensor))
    return x.detach() if isinstance(x, torch.Tensor) else torch.from_numpy(x)


def save_tensor_to_json(data: Union[Tensor, np.ndarray], path: PathLike):
    "Save tensor to disk as a JSON file"
    data = to_np(data) if isinstance(data, Tensor) else data
    if not isinstance(data, (Tensor, np.ndarray)):
        raise ValueError(f"Invalid `data` type. Expected tensor/array, got {type(data)}")
    data = data.tolist()

    path = Path(path).with_suffix(".json")
    # path = Path(path)
    # filename = f"{path.stem}.json"  # Ensure extension is .json
    # path = path.parent / filename
    write_json(data, path, indent=None)



from upyog.ml.imports import *


__all__ = ["compare_state_dicts", "set_bn_eval"]


ModuleOrWeights = Union[nn.Module, OrderedDict]

def compare_state_dicts(
    m1: ModuleOrWeights, m2: ModuleOrWeights, verbose=True, name: str = "models"
) -> bool:
    mismatch_keys = []
    m1 = m1.state_dict() if isinstance(m1, nn.Module) else m1
    m2 = m2.state_dict() if isinstance(m2, nn.Module) else m2

    for (k1, m1v), (k2, m2v) in zip(m1.items(), m2.items()):
        assert k1 == k2, f"Key mismatch, looks like you're comparing different models"
        if not torch.equal(m1v.cpu(), m2v.cpu()):
            logger.error(
                f"Mismatch in <fg #0F77A9>{name}</> in layer <fg #0F77A9>{k1}</>"
            )
            mismatch_keys.append(k1)

    if not mismatch_keys == []:
        err_msg = f"Weights didn't match exactly in {len(mismatch_keys)} layers: {mismatch_keys}"
        raise RuntimeError(err_msg if verbose else "")
    logger.success(f"`state_dict`s of both <fg #0F77A9>{name} </> match!")
    return True


BN_LAYER_TYPES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)
def set_bn_eval(m: nn.Module) -> None:
    "Set bn layers in eval mode for all recursive children of `m`."
    for l in m.children():
        if isinstance(l, BN_LAYER_TYPES) and not next(l.parameters()).requires_grad:
            l.eval()
        set_bn_eval(l)



from upyog.ml.imports import *
import coremltools as ct


__all__ = ["get_coreml_input"]


# TODO: Get from upyog
def get_coreml_input(
    H: int,
    W: int,
    norm_mean: Optional[Tuple[float, float, float]] = (0.485, 0.456, 0.406),
    norm_stdev: Optional[Tuple[float, float, float]] = (0.229, 0.224, 0.225),
    input_name: str = "Image",
    input_format: Literal["RGB", "BGR"] = "RGB",
) -> List[ct.ImageType]:
    """
    Get CoreML input image type to feed into a model. If using `norm_mean` and
    `norm_stdev`, it does preprocessing as described (ImageNet example) here:
    https://github.com/apple/coremltools/issues/1160

    If `norm_mean` or `norm_stdev` is None, inputs are are not normalised
    and are in the range of 0-255 (as is the case for YOLOX)

    Returns a list so you can call the function inside the converter alone
        `ct.convert(model, inputs = get_coreml_input(...))`
    """
    if norm_mean and norm_stdev:
        assert len(norm_mean) == 3 and len(norm_stdev) == 3
        scale_div = sum(norm_stdev) / len(norm_stdev)
        scale = 1.0 / (255.0 * scale_div)
        bias = list(map(lambda x: -x / scale_div, norm_mean))
    else:
        scale = 1.0
        bias = None

    input_type = ct.ImageType(
        input_name,
        shape=torch.rand(1, 3, H, W).shape,
        scale=scale,
        bias=bias,
        color_layout=input_format,
    )
    return [input_type]


from upyog.ml.imports import *
from upyog.image import fig_to_pil


__all__ = ["visualise_conf_matrix"]


def visualise_conf_matrix(
    conf_matrix: np.ndarray,
    labels: List[str],
    title: Optional[str] = None,
    normalize: bool = True,
    cmap: Union[None, str, mpl.colors.Colormap] = None,
    width: int = 6,
    dpi: int = 200,
    return_pil: bool = True,
):
    # Convert to DataFrame
    cm = conf_matrix
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        cm = pd.DataFrame(cm * 100)
    else:
        cm = pd.DataFrame(cm)
    cm.index = labels
    cm.columns = labels

    # Setup plot canvas
    fig, ax = plt.subplots(figsize=(width, width), dpi=dpi)

    # Colormap
    if not cmap:
        cmap = 'Blues'
        # cmap = sns.cubehelix_palette(as_cmap=True)

    # Draw heatmap, axes and title
    sns.heatmap(cm, annot=True, fmt=".1f", cbar=False, ax=ax, cmap=cmap)
    _= ax.set_yticklabels(labels, rotation=0)
    _= ax.set_xticklabels(labels, rotation=90)

    if title:
        ax.set_title(title)

    if return_pil: return fig_to_pil(fig)
    else:          return fig
